/**************************************************************************
*  Copyright (c) 2013 by Michael Fischer. All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  09.02.2013  mifi  First Version for the BeagleBone board.
*  14.02.2013  mifi  Corrected MMUConfigAndEnable, DEV region.
*                    Use MMU_MEMTYPE_STRONG_ORD_SHAREABLE instead of
*                    MMU_MEMTYPE_DEVICE_SHAREABLE.
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/

#include <stdint.h>
#include <ctl_api.h>
#include <libarm.h>
#include "ch.h"

#include "mmu.h"
#include "cache.h"
#include "cp15.h"
#include "dmtimer.h"
#include "interrupt.h"

#include "soc_AM335x.h"
#include "hw_control_AM335x.h"
#include "hw_types.h"
#include "hw_cm_per.h"
#include "hw_cm_dpll.h"
#include "hw_gpio_v2.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*
 * Some defines for the MMU translation table
 */
#define START_ADDR_DDR                (0x80000000)
#define START_ADDR_DEV                (0x44000000)
#define START_ADDR_OCMC               (0x40300000)
#define NUM_SECTIONS_DDR              (512)
#define NUM_SECTIONS_DEV              (960)
#define NUM_SECTIONS_OCMC             (1)


/*
 * Some defines for the pin mux setup
 */
#define CONTROL_CONF_MMODE_SHIFT       0UL
#define CONTROL_CONF_PUDEN_SHIFT       3UL
#define CONTROL_CONF_PUTYPESEL_SHIFT   4UL
#define CONTROL_CONF_RXACTIVE_SHIFT    5UL


/*
 * Priority and stack size of the task.
 */
#define TASK_HP_PRIORITY   NORMALPRIO + 3
#define TASK_LP_PRIORITY   NORMALPRIO + 2

#define TASK_HP_STK_SIZE   128
#define TASK_LP_STK_SIZE   128

/*=======================================================================*/
/*  Prototypes                                                           */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static volatile unsigned int pageTable[4*1024] __attribute__((aligned(16*1024)));

static WORKING_AREA(HPStack, TASK_HP_STK_SIZE);
static WORKING_AREA(LPStack, TASK_LP_STK_SIZE);

static Thread* pHPThread;
static Thread* pLPThread;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  TimerTickIrqHandler                                                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static CH_IRQ_HANDLER (TimerTickIrqHandler) 
{
  CH_IRQ_PROLOGUE();

  chSysLockFromIsr();
  chSysTimerHandlerI();
  chSysUnlockFromIsr();
   
  /* Clear the status of the interrupt flags */
  DMTimerIntStatusClear(SOC_DMTIMER_2_REGS, DMTIMER_INT_OVF_IT_FLAG);
   
  CH_IRQ_EPILOGUE();
} /* TimerTickIrqHandler */

/*************************************************************************/
/*  InitTimerTick                                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void InitTimerTick (void)
{
  uint32_t counter;
  
  /* 
   * Select the clock source for the Timer2 instance. 
   * (Based on the DMTimer2ModuleClkConfig function from the BeagleBone)
   */
  HWREG(SOC_CM_DPLL_REGS + CM_DPLL_CLKSEL_TIMER2_CLK) &=
        ~(CM_DPLL_CLKSEL_TIMER2_CLK_CLKSEL);

  HWREG(SOC_CM_DPLL_REGS + CM_DPLL_CLKSEL_TIMER2_CLK) |=
        CM_DPLL_CLKSEL_TIMER2_CLK_CLKSEL_CLK_M_OSC;

  while((HWREG(SOC_CM_DPLL_REGS + CM_DPLL_CLKSEL_TIMER2_CLK) &
         CM_DPLL_CLKSEL_TIMER2_CLK_CLKSEL) !=
         CM_DPLL_CLKSEL_TIMER2_CLK_CLKSEL_CLK_M_OSC);

  /* Writing to MODULEMODE field of CM_PER_TIMER2_CLKCTRL register. */
  HWREG(SOC_CM_PER_REGS + CM_PER_TIMER2_CLKCTRL) |=
        CM_PER_TIMER2_CLKCTRL_MODULEMODE_ENABLE;

  /* Waiting for MODULEMODE field to reflect the written value. */
  while((HWREG(SOC_CM_PER_REGS + CM_PER_TIMER2_CLKCTRL) &
         CM_PER_TIMER2_CLKCTRL_MODULEMODE) != 
         CM_PER_TIMER2_CLKCTRL_MODULEMODE_ENABLE);

  /*
   * Waiting for IDLEST field in CM_PER_TIMER2_CLKCTRL register 
   * for the module is fully functional.
   */
  while((HWREG(SOC_CM_PER_REGS + CM_PER_TIMER2_CLKCTRL) & 
         CM_PER_TIMER2_CLKCTRL_IDLEST) != 
         CM_PER_TIMER2_CLKCTRL_IDLEST_FUNC);
   
  /* Waiting for the L4LS and TIMER2 clock */
  while(!(HWREG(SOC_CM_PER_REGS + CM_PER_L4LS_CLKSTCTRL) &
         (CM_PER_L4LS_CLKSTCTRL_CLKACTIVITY_L4LS_GCLK |
          CM_PER_L4LS_CLKSTCTRL_CLKACTIVITY_TIMER2_GCLK)));


  /*
   * Setup timer.
   */
   
  /* Registering DMTimerIsr */
  IntRegister(SYS_INT_TINT2, TimerTickIrqHandler);

  /* Set the priority */
  IntPrioritySet(SYS_INT_TINT2, 0x3e, AINTC_HOSTINT_ROUTE_IRQ);

  /* Enable the system interrupt */
  IntSystemEnable(SYS_INT_TINT2);

  /* Setup timer */
  counter = 0xffffffff - (24000000UL / CH_FREQUENCY);
  
  DMTimerReset(SOC_DMTIMER_2_REGS);
  DMTimerCounterSet(SOC_DMTIMER_2_REGS, counter);
  DMTimerReloadSet(SOC_DMTIMER_2_REGS, counter);
  DMTimerModeConfigure(SOC_DMTIMER_2_REGS, DMTIMER_AUTORLD_NOCMP_ENABLE);
  
  /* Enable timer interrupt */
  DMTimerIntEnable(SOC_DMTIMER_2_REGS, DMTIMER_INT_OVF_EN_FLAG);
  
  /* Start timer */
  DMTimerEnable(SOC_DMTIMER_2_REGS);   

} /* InitTimerTick */

/*************************************************************************/
/*  MMUConfigAndEnable                                                   */
/*                                                                       */
/*  Function to setup MMU. This function Maps three regions              */
/*  1. DDR, 2. OCMC and 3. Device memory, and enables MMU.               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void MMUConfigAndEnable (void)
{
  /* Disable cache first */
  CacheDisable(CACHE_ALL);

  /*
  ** Define DDR memory region of AM335x. DDR can be configured as Normal
  ** memory with R/W access in user/privileged modes. The cache attributes
  ** specified here are,
  ** Inner - Write through, No Write Allocate
  ** Outer - Write Back, Write Allocate
  */
  REGION regionDdr = {
                      MMU_PGTYPE_SECTION, START_ADDR_DDR, NUM_SECTIONS_DDR,
                      MMU_MEMTYPE_NORMAL_NON_SHAREABLE(MMU_CACHE_WT_NOWA,
                                                       MMU_CACHE_WB_WA),
                      MMU_REGION_NON_SECURE, MMU_AP_PRV_RW_USR_RW,
                      (unsigned int*)pageTable
                     };
  /*
  ** Define OCMC RAM region of AM335x. Same Attributes of DDR region given.
  */
  REGION regionOcmc = {
                       MMU_PGTYPE_SECTION, START_ADDR_OCMC, NUM_SECTIONS_OCMC,
                       MMU_MEMTYPE_NORMAL_NON_SHAREABLE(MMU_CACHE_WT_NOWA,
                                                        MMU_CACHE_WB_WA),
                       MMU_REGION_NON_SECURE, MMU_AP_PRV_RW_USR_RW,
                       (unsigned int*)pageTable
                      };

  /*
  ** Define Device Memory Region. The region between OCMC and DDR is
  ** configured as device memory, with R/W access in user/privileged modes.
  ** Also, the region is marked 'Execute Never'.
  */
  REGION regionDev = {
                      MMU_PGTYPE_SECTION, START_ADDR_DEV, NUM_SECTIONS_DEV,
                      MMU_MEMTYPE_STRONG_ORD_SHAREABLE,
                      MMU_REGION_NON_SECURE,
                      MMU_AP_PRV_RW_USR_RW  | MMU_SECTION_EXEC_NEVER,
                      (unsigned int*)pageTable
                     };

  /* Initialize the page table and MMU */
  MMUInit((unsigned int*)pageTable);

  /* Map the defined regions */
  MMUMemRegionMap(&regionDdr);
  MMUMemRegionMap(&regionOcmc);
  MMUMemRegionMap(&regionDev);

  /* Now Safe to enable MMU */
  MMUEnable((unsigned int*)pageTable);

  /* Enable the cache too */
  CacheEnable(CACHE_ALL);

} /* MMUConfigAndEnable */

/*************************************************************************/
/*  HardwareInit                                                         */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void HardwareInit (void)
{
  uint32_t Value;
  
  /*
   * Setup cache and mmu
   */  
  MMUConfigAndEnable();

    
  /*
   * Initialize the ARM interrupt control 
   */
  IntAINTCInit();
  

  /*
   * Setup the SysTick timer
   */
  InitTimerTick();   


  /*
   * Enable GPIO for the LEDs
   */
   
  /* Enable the GPIO1 module */
  HWREG(SOC_PRCM_REGS + CM_PER_GPIO1_CLKCTRL) = 
        CM_PER_GPIO1_CLKCTRL_OPTFCLKEN_GPIO_1_GDBCLK |
        CM_PER_GPIO1_CLKCTRL_MODULEMODE_ENABLE;
        
  /*
   * Waiting for IDLEST field in CM_PER_GPIO1_CLKCTRL register 
   * for the module is fully functional.
   */
  while((HWREG(SOC_CM_PER_REGS + CM_PER_GPIO1_CLKCTRL) & 
         CM_PER_GPIO1_CLKCTRL_IDLEST) != 
         CM_PER_GPIO1_CLKCTRL_IDLEST_FUNC);
        
        
  /*
   * Setup Pad Control Registers
   */
  Value = ( 7 << CONTROL_CONF_MMODE_SHIFT )     |    /* Mode 7 */
          ( 1 << CONTROL_CONF_PUDEN_SHIFT )     |    /* Pull disabled */
          ( 1 << CONTROL_CONF_PUTYPESEL_SHIFT ) |    /* Pullup */
          ( 0 << CONTROL_CONF_RXACTIVE_SHIFT );      /* Receiver disabled */
  
  HWREG(SOC_CONTROL_REGS + CONTROL_CONF_GPMC_AD(7)) = Value;
  
  /* Module is enabled, clocks are not gated. */
  HWREG(SOC_GPIO_1_REGS + GPIO_CTRL) &= ~GPIO_CTRL_DISABLEMODULE;

  /* Set the data */
  HWREG(SOC_GPIO_1_REGS + GPIO_SETDATAOUT) = (1 << 7);
  
  /* The corresponding GPIO port is configured as an output. */
  HWREG(SOC_GPIO_1_REGS + GPIO_OE) &= ~(1 << 7);
  
} /* HardwareInit */

/***************************************************************************/
/*  LedOn                                                                  */
/*                                                                         */
/*  In    : none                                                           */
/*  Out   : none                                                           */
/*  Return: none                                                           */
/***************************************************************************/
static void LedOn (void)
{
  HWREG(SOC_GPIO_1_REGS + GPIO_CLEARDATAOUT) = (1 << 7);
} /* LedOn */

/***************************************************************************/
/*  LedOff                                                                 */
/*                                                                         */
/*  In    : none                                                           */
/*  Out   : none                                                           */
/*  Return: none                                                           */
/***************************************************************************/
static void LedOff (void)
{
  HWREG(SOC_GPIO_1_REGS + GPIO_SETDATAOUT) = (1 << 7);
} /* LedOff */

/***************************************************************************/
/*  HPTask                                                                 */
/*                                                                         */
/*  In    : task parameter                                                 */
/*  Out   : none                                                           */
/*  Return: never                                                          */
/***************************************************************************/
static msg_t HPTask (void *p)
{
  Thread *tp;
  msg_t Message;
  p = p;
  
  while(1)
  {
    tp = chMsgWait();
    Message = chMsgGet(tp);
    chMsgRelease(tp, Message);
    LedOff();
  }
} /* HPTask */

/***************************************************************************/
/*  LPTask                                                                 */
/*                                                                         */
/*  In    : task parameter                                                 */
/*  Out   : none                                                           */
/*  Return: never                                                          */
/***************************************************************************/
static msg_t LPTask (void *p)
{
  msg_t LPmessage = 42;
  
  p = p;
  
  while(1)
  {
    chThdSleepMilliseconds(100);
     
    LedOn();
    LedOff();
     
    LedOn();
    chMsgSend(pHPThread, 42);
  }
} /* LPTask */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*************************************************************************/
int main (void)
{
  /*
   * Before the application which is using ChibiOS/RT can be started, 
   * the hardware of the target must be initialized. The CPU is still
   * initialized by the CrossWorks for ARM startup. But the timer tick
   * must be init too. Even the GPIO can be init here.
   */
  HardwareInit();
  
  /*
   * ChibiOS/RT initialization.
   */
  chSysInit();
  
  /*
   * Create our threads
   */
  pHPThread = chThdCreateStatic(HPStack, sizeof(HPStack), TASK_HP_PRIORITY, HPTask, NULL);
  pLPThread = chThdCreateStatic(LPStack, sizeof(LPStack), TASK_LP_PRIORITY, LPTask, NULL);
  
  /*
   * Normal main() thread activity.
   */
  while(1) 
  {
    chThdSleepMilliseconds(500);
  }

  /*
   * This return here make no sense.
   * But to prevent the compiler warning:
   * "return type of 'main' is not 'int'
   * we use an int as return :O)
   */ 
  return(0);
} /* main */

/*** EOF ***/
